import ctypes
from time import sleep

from mag_tools.log.logger import Logger
from mag_tools.model.convert_type import ConvertType
from mag_tools.model.log_type import LogType
from mag_tools.utils.common.time_probe import TimeProbe
from selenium.common.exceptions import InvalidSelectorException, NoSuchElementException, TimeoutException, \
    WebDriverException
from selenium.webdriver import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.support.wait import WebDriverWait

from mag_test.bean.element_info import ElementInfo
from mag_test.core.app_driver import AppDriver
from mag_test.finder.driver_finder_utils import DriverFinderUtils
from mag_test.finder.element_finder_utils import ElementFinderUtils
from mag_test.finder.window_finder import WindowFinder
from mag_test.model.control_type import ControlType
from mag_test.utils.tree_utils import TreeUtils
from mag_test.utils.event_utils import EventUtils


class ElementFinder:
    @staticmethod
    def find(driver:AppDriver, element_info:ElementInfo):
        Logger.debug(f'开始查找控件：{element_info.self_info()}')
        probe = TimeProbe.get_probe('查找控件')

        parent = ElementFinder.__find_parent(driver, element_info)

        # 按钮/拆分按钮
        if element_info.control_type in {ControlType.BUTTON, ControlType.SPLIT_BUTTON}:
            element = ElementFinder.__find_button(driver, element_info, parent)
        # 菜单
        elif element_info.control_type == ControlType.MENU:
            element = ElementFinder.__find_menu(driver, element_info, parent)
        # 组合框
        elif element_info.control_type == ControlType.COMBO_BOX:
            element = ElementFinder.__find_combox_box(driver, element_info, parent)
        # 列表框
        elif element_info.control_type == ControlType.LIST:
            element = ElementFinder.__find_list_box(driver, element_info, parent)
        # 列表视图
        elif element_info.control_type == ControlType.LIST_VIEW:
            element = ElementFinder.__find_list_view(driver, element_info, parent)
        # 树视图
        elif element_info.control_type == ControlType.TREE:
            element = ElementFinder.__find_tree_view(driver, element_info, parent)
        # 表格
        elif element_info.control_type == ControlType.TABLE:
            element = ElementFinder.__find_table(driver, element_info, parent)
        # 日期时间
        elif element_info.control_type == ControlType.DATETIME:
            element = ElementFinder.__find_datetime(driver, element_info, parent)
        # 窗口
        elif element_info.control_type == ControlType.WINDOW:
            element = ElementFinder.__find_window(driver, element_info)
        # 简单控件（包括：工具栏、Pane等）
        else:
            print('simple in')

            element = ElementFinder.__find_main_element(driver, element_info, parent)

        probe.write_log()

        return element

    @staticmethod
    def switch_to_window_by_title(driver:AppDriver, window_title:str):
        window_title = window_title.strip()

        if '.exe' in window_title.lower():
            driver = driver.new_driver(window_title)
        else:
            try:
                hwnd_list = WebDriverWait(driver, 10).until(lambda drv: WindowFinder.find_window_by_title(window_title))
                if hwnd_list:
                    hwnd = hwnd_list[0]
                    ctypes.windll.user32.SetForegroundWindow(hwnd)
                else:
                    Logger.error(LogType.FRAME, f"未找到窗口: {window_title}")
            except InvalidSelectorException as e:
                Logger.error(LogType.FRAME, f"查找窗口[{window_title}]选择器无效: {str(e)}")
            except NoSuchElementException as e:
                Logger.error(LogType.FRAME, f"查找窗口[{window_title}]出错: {str(e)}")
            except TimeoutException as e:
                Logger.error(LogType.FRAME, f"查找窗口[{window_title}]超时: {str(e)}")
            except WebDriverException as e:
                Logger.error(LogType.FRAME, f"查找窗口[{window_title}]时通讯异常: {str(e)}")
            except Exception as e:
                Logger.error(LogType.FRAME, f"未知异常：{str(e)}")

        return driver

    @staticmethod
    def __find_button(driver, element_info, parent=None):
        """
        查找按钮控件（包括：普通按钮和拆分按钮）
        参数：
        name 按钮名
        parent 父控件
        """
        for element_type in [ControlType.BUTTON, ControlType.SPLIT_BUTTON]:
            element_info.control_type = element_type
            element = ElementFinder.__find_main_element(driver, element_info, parent)
            if element is not None:
                break
        return element

    @staticmethod
    def __find_menu(driver, element_info, parent=None):
        """
        查找菜单控件
        参数：
        name 菜单及菜单项名，格式：菜单项名/子菜单项名/...
        parent_name 父控件名
        parent_control_type 父控件类型
        """
        items = element_info.get_menu_items()

        element = None
        actions = ActionChains(driver)

        if parent:
            for index, item in enumerate(items):
                element = ElementFinderUtils.find_element_by_type(element, item, ControlType.MENU_ITEM)
                if index < len(items) - 1:
                    actions.move_to_element(element).click().perform()
        else:
            for index, item in enumerate(items):
                element = DriverFinderUtils.find_element_by_type_wait(driver, item, ControlType.MENU_ITEM)
                if index < len(items) - 1:
                    actions.move_to_element(element).click().perform()
        return element

    @staticmethod
    def __find_combox_box(driver, element_info, parent=None):
        """
            查找组合框或列表项
            参数：
            name 组合框及列表项名，格式：组合框名/列表项名/，列表项名可为空
        """
        combo_box = ElementFinder.__find_main_element(driver, element_info, parent)

        if element_info.child_name:
            combo_box.click()
            return ElementFinderUtils.find_element_by_type(combo_box, element_info.child_name, ControlType.LIST_ITEM)
        else:
            return combo_box

    @staticmethod
    def __find_list_box(driver, element_info, parent=None):
        """
        查找列表框
        参数：
        name 列表框名，格式：列表框名/列表项名/菜单名，列表框名和菜单名可为空，只支持简单模式的弹出菜单
        """
        list_box = ElementFinder.__find_main_element(driver, element_info, parent)

        list_box.click()
        element = ElementFinderUtils.find_element_by_type(list_box, element_info.child_name, ControlType.LIST_ITEM)
        element = ElementFinder.__find_context_menu(driver, element, element_info.menu_item)

        return element

    @staticmethod
    def __find_list_view(driver, element_info, parent=None):
        """
            查找列表视图
            参数：
            name 列表视图名，格式：列表视图名/列表项序号/菜单名，列表视图名和菜单名可为空，只支持菜单模式的弹出菜单
        """

        list_view = ElementFinder.__find_main_element(driver, element_info, parent)

        element = ElementFinderUtils.find_element_by_type(list_view, element_info.child_name, ControlType.LIST_ITEM)
        element = ElementFinder.__find_context_menu(driver, element, element_info.menu_item)
        return element

    @staticmethod
    def __find_tree_view(driver, element_info, parent=None):
        """
            查找树视图
            参数：
            name 树视图名，格式：树视图名/树节点名/菜单名，树视图名和菜单名可为空，只支持菜单模式的弹出菜单
        """
        tree_view = ElementFinder.__find_main_element(driver, element_info, parent)

        if element_info.child_name:
            TreeUtils.expand_all(driver, tree_view)
            element = ElementFinderUtils.find_element_by_type(tree_view, element_info.child_name, ControlType.TREE_ITEM)
        else:
            element = tree_view

        if element_info.menu_item:
            element = ElementFinder.__find_context_menu(driver, element, element_info.menu_item)

        if element is not None:
            Logger.debug(LogType.FRAME, f"树或弹出菜单项为：{element}")

        return element

    @staticmethod
    def __find_table(driver, element_info, parent=None):
        """
            查找表格
            参数：
            name 表格名，格式：表格名/文本框名/菜单名，表格名和菜单名可为空，只支持菜单模式的弹出菜单
            返回：TableRow的数组或Edit
        """
        table = ElementFinder.__find_main_element(driver, element_info, parent)

        if element_info.child_name and ConvertType.of_code(element_info.child_name) is None:
            element = None

            table_rows = ElementFinderUtils.find_elements_by_type(table, None, ControlType.TABLE_ROW)
            for row_index, row in enumerate(table_rows):
                cells = ElementFinderUtils.find_elements_by_type(row, None, ControlType.EDIT)
                for cell_index, cell in enumerate(cells):
                    if cell.text == element_info.child_name:
                        element = cell
                        break
        else:
            element = table

        if element_info.menu_item:
            element = ElementFinder.__find_context_menu(driver, element, element_info.menu_item)

        if element is not None:
            Logger.debug(LogType.FRAME, f"表格或弹出菜单项为：{element}")

        return element
    
    @staticmethod
    def __find_datetime(driver, element_info, parent=None):
        """
        查找日期时间控件
        参数：
        name 日期时间控件名
        """
        if parent:
            dt = ElementFinderUtils.find_element_by_class(parent, element_info.name, 'SysDateTimePick32')
        else:
            dt = DriverFinderUtils.find_element_by_class(driver, element_info.name, 'SysDateTimePick32')

        return dt

    @staticmethod
    def __find_window(driver, element_info):
        """
        查找窗口
        参数：
        name 窗口名（关键词）
        """
        return driver.find_element(By.XPATH, f"//Window[contains(@Name, '{element_info.name}')]")
    
    @staticmethod
    def __find_context_menu(driver, element, menu_item_name):
        if menu_item_name:
            actions = ActionChains(driver)
            actions.move_to_element(element)
            actions.context_click(element).perform()

            try:
                menu = DriverFinderUtils.find_element_by_class(driver.root_driver, '上下文', '#32768')
                element = ElementFinderUtils.find_element_by_type(menu, menu_item_name, ControlType.MENU_ITEM)
            except (NoSuchElementException, InvalidSelectorException, WebDriverException) :
                element = DriverFinderUtils.find_element_by_type(driver, menu_item_name, ControlType.MENU_ITEM)

        return element

    @staticmethod
    def __find_main_element(driver, element_info, parent=None):
        try:
            if element_info.automation_id:
                if parent:
                    element = ElementFinderUtils.find_element_by_automation(parent, element_info.automation_id)
                else:
                    element = DriverFinderUtils.find_element_by_automation_wait(driver, element_info.automation_id)
            else:
                if parent:
                    element = ElementFinderUtils.find_element_by_type(parent, element_info.name, element_info.control_type)
                else:
                    element = DriverFinderUtils.find_element_by_type_wait(driver, element_info.name, element_info.control_type)

            if element_info.control_type.is_composite():
                offset = element_info.get_parent_offset(element.size['width'], element.size['height'])
                EventUtils.click(driver, element, offset)
        except (NoSuchElementException, InvalidSelectorException, WebDriverException, TimeoutException) as e:
            element = None
            Logger.debug(f'未找到主控件：{str(e)}')    # 模糊查找控件时，返回None正常

        return element

    @staticmethod
    def __find_parent(driver, element_info):
        parent = None
        if element_info.need_to_find_parent():
            try:
                Logger.debug(f'开始查找父控件：{element_info.parent_info()}')
                probe = TimeProbe.get_probe('查找父控件')

                if element_info.parent_id:
                    parent = DriverFinderUtils.find_element_by_automation_wait(driver, element_info.parent_id)
                elif element_info.parent_type:
                    parent = DriverFinderUtils.find_element_by_type_wait(driver, element_info.parent_name,
                                                                   element_info.parent_type)

                probe.write_log()
            except (NoSuchElementException, InvalidSelectorException, WebDriverException) as e:
                Logger.error(f'查找父控件失败：{element_info.parent_info()}')
                raise e

        return parent


import logging
import threading
import time
from typing import Dict, List, Optional, Union

from appium import webdriver
from appium.webdriver import WebElement
from mag_tools.log.logger import Logger
from mag_tools.model.log_type import LogType
from mag_tools.model.message_type import MessageType
from mag_tools.utils.common.message_box import MessageBox
from mag_tools.model.control_type import ControlType
from selenium.common.exceptions import WebDriverException, SessionNotCreatedException, InvalidArgumentException, \
    TimeoutException
from selenium.webdriver.common.by import By

from mag_tools.utils.common.process_utils import ProcessUtils
from mag_tools.utils.test.element_finder_utils import ElementFinderUtils


class AppDriver(webdriver.Remote):
    def __init__(self, url:str, capabilities:Dict[str,str], parent_driver:webdriver.Remote=None):
        logging.basicConfig(level=logging.DEBUG)

        self.app = capabilities["app"]
        self.parent_driver = parent_driver
        self.__stop_event = threading.Event()
        self.__url = url
        self.__capabilities = capabilities

        # 检查应用程序是否已经在运行
        proc = ProcessUtils.find_app(self.app)
        if proc:
            proc.terminate()

        try:
            self.root_driver = webdriver.Remote(command_executor=url,
                                            desired_capabilities={"platformName": "Windows",
                                                                  "deviceName": "WindowsPC",
                                                                  "app": "Root"})
        except SessionNotCreatedException:
            MessageBox.showinfo("会话未建立", "检查WinAppDriver是否已启动及网络是否正常！")
        except InvalidArgumentException:
            MessageBox.showinfo("参数无效", "检查参数是否正确！")
        except TimeoutException:
            MessageBox.showinfo("连接超时", "，检查WinAppDriver是否已启动及网络是否正常！")
        except WebDriverException:
            MessageBox.showinfo("连接出错", "检查WinAppDriver是否已启动及网络是否正常！")
        except Exception as e:
            MessageBox.showinfo("未知异常", str(e))

        # 初始化webdriver
        super().__init__(command_executor=url, desired_capabilities=capabilities)

    def new_driver(self, app):
        return AppDriver(self.__url, self.__get_capabilities(app), self)

    def check_alert(self)->List[Union[Optional[MessageType], Optional[str]]]:
        alert_result = self.__find_alert()
        if alert_result[0] is not None:
            alert_result[0].click()
        return alert_result[1:]

    def stop_checking_for_alert(self):
        self.__stop_event.set()

    def quit_app(self):
        try:
            if self.session_id:
                time.sleep(2)
                self.root_driver.quit()
                time.sleep(2)
                self.quit()
                Logger.info(LogType.FRAME, f"关闭'{self.app}' driver完毕")
        except AttributeError:
            Logger.error(LogType.FRAME, "No active session to quit")

    def __find_alert(self) -> List[Union[Optional[WebElement], Optional[MessageType], Optional[str]]]:
        alert_result = [None, None, None]
        try:
            exp = ElementFinderUtils.global_expression(None, None, '#32770', None, None, None)
            alerts = self.find_elements(By.XPATH, exp)
            if alerts is not None and len(alerts) > 0:
                alert = alerts[-1] # 获取最新弹出的对话框
                alert_title = alert.get_attribute("Name")

                keyword_found = next((keyword for keyword in ["警告", "错误", "确认", "HiSim"] if keyword in alert_title), None)
                if keyword_found is not None:
                    message_type = MessageType.of_code(keyword_found) if keyword_found else None

                    alert_label = alert.find_element(By.XPATH, f".//{ControlType.LABEL.code}")
                    message = alert_label.get_attribute("Name")
                    Logger.info(LogType.SERVICE, f'message: {message}')

                    element = alert.find_element(By.XPATH, f".//{ControlType.BUTTON.code}[@Name='确定(O)' or @Name='是(Y)' or @Name='确定']")

                    alert_result = [element, message_type, message]
        except Exception as e:
            Logger.error(LogType.FRAME, f"Exception: {str(e)}")

        return alert_result

    def __get_capabilities(self, app):
        return {
            'platformName': self.__capabilities['platformName'],
            'deviceName': self.__capabilities['deviceName'],
            'automationName': self.__capabilities['automationName'],
            'app': app,
            'appWorkingDir': self.__capabilities['appWorkingDir'],
        }


import json
import os

import pytest
from mag_test.model.function_status import FunctionStatus
from mag_test.model.test_component_type import TestComponentType
from mag_test.model.control_type import ControlType

from mag_test.bean.base_test import BaseTest
from mag_test.bean.step import Step
from mag_test.core.app_driver import AppDriver

class Function(BaseTest):
    def __init__(self, home_dir:str, plan_id:str, function_id:str, index:int, name:str, status:FunctionStatus=FunctionStatus.NORMAL):
        super().__init__(home_dir, name, index, TestComponentType.FUNCTION, None)
        self.__plan_id = plan_id
        self.__id = function_id  # 功能标识
        self.__status = status if status else FunctionStatus.NORMAL
        self.__steps = []

        self.__read()

    @pytest.mark.benchmark
    def start(self, driver:AppDriver)->AppDriver:
        """
        启动测试功能
        :param driver: AppDriver
        """
        if self.__status == FunctionStatus.NORMAL:
            test_failed = False
            for step in self.__steps:
                if test_failed:
                    step.skip()
                else:
                    driver = step.start(driver)
                    if step.is_fail():
                        self.fail('该功能测试失败')
                        test_failed = True

            super()._report()
        return driver

    def __read(self):
        function_file = os.path.join(self.script_dir, self.__plan_id, f'{self.__id}.json')
        with open(function_file, 'r', encoding='utf-8') as file:
            data = json.load(file)
            if data['name']:
                self.__name = data['name']
                self.__status = FunctionStatus[data['status']]

            for index, item in enumerate(data.get('elements', []), start=1):
                step = Step(self._home_dir,
                            item.get("step", None),
                            item.get("control_name", None),
                            ControlType[item.get('control_type')] if item.get('control_type') else None,
                            item.get('id', None),
                            item.get('value', None),
                            self._index,
                            index,
                            item.get('parent', None),
                            ControlType[item.get('parent_type')] if item.get('parent_type') else None,
                            item.get('parent_id', None),
                            item.get('pop', None))
                self.__steps.append(step)


from selenium.webdriver import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC

from mag_test.core.app_driver import AppDriver
from mag_test.model.control_type import ControlType
from mag_test.finder.element_finder_utils import ElementFinderUtils


class TreeUtils:
    @staticmethod
    def expand_all(driver:AppDriver, tree:WebElement):
        try:
            actionChains = ActionChains(driver)

            # 查找所有展开按钮
            tree_items = ElementFinderUtils.find_elements_by_type(tree, None, ControlType.TREE_ITEM)

            for tree_item in tree_items:
                # 双击展开
                actionChains.double_click(tree_item).perform()
                # # 等待子节点加载完成
                # WebDriverWait(driver, 10).until(
                #     EC.presence_of_element_located((By.XPATH, './/Button[@content-desc="Expand"]')))
                # 递归展开子节点
                # TreeUtils.expand_all(driver, tree)
        except Exception as e:
            print(f"Error expanding nodes: {str(e)}")